@@ -28,11 +28,8 @@ type pat =
   | PInac of constr
   | PHide of int
 
-type context_map = {
-  src_ctx : rel_context;
-  map_inst : pat list;
-  tgt_ctx : rel_context;
-}
+type context_map = rel_context * pat list * rel_context
+
 
 let type_of_refresh env evdref c =
   let ty = Retyping.get_type_of env !evdref c in
@@ -97,20 +94,20 @@ and pats_vars l =
 
 let inaccs_of_constrs l = List.map (fun x -> PInac x) l
 
-let rec pats_of_constrs env sigma l = List.map (pat_of_constr env sigma) l
-and pat_of_constr env sigma c =
+let rec pats_of_constrs sigma l = List.map (pat_of_constr sigma) l
+and pat_of_constr sigma c =
   match kind sigma c with
   | Rel i -> PRel i
-  | App (f, [| a ; c |]) when is_global env sigma (Lazy.force coq_inacc) f ->
+  | App (f, [| a ; c |]) when is_global sigma (Lazy.force coq_inacc) f ->
     PInac c
-  | App (f, [| a ; c |]) when is_global env sigma (Lazy.force coq_hide) f ->
+  | App (f, [| a ; c |]) when is_global sigma (Lazy.force coq_hide) f ->
     PHide (destRel sigma c)
   | App (f, args) when isConstruct sigma f ->
     let ((ind,_),_ as cstr) = destConstruct sigma f in
     let nparams = Inductiveops.inductive_nparams (Global.env()) ind in
     let params, args = Array.chop nparams args in
     PCstr (cstr, inaccs_of_constrs (Array.to_list params) @
-                 pats_of_constrs env sigma (Array.to_list args))
+                 pats_of_constrs sigma (Array.to_list args))
   | Construct f -> PCstr (f, [])
   | _ -> PInac c
 
@@ -134,9 +131,9 @@ let rec pat_to_user_pat ?(avoid = ref Id.Set.empty) ?loc ctx = function
 and pats_to_lhs ?(avoid = ref Id.Set.empty) ?loc ctx pats =
   List.map_filter (pat_to_user_pat ~avoid ?loc ctx) pats
 
-let context_map_to_lhs ?(avoid = Id.Set.empty) ?loc map =
+let context_map_to_lhs ?(avoid = Id.Set.empty) ?loc (ctx, pats, _) =
   let avoid = ref avoid in
-  List.rev (pats_to_lhs ~avoid ?loc map.src_ctx map.map_inst)
+  List.rev (pats_to_lhs ~avoid ?loc ctx pats)
 
 let do_renamings env sigma ctx =
   let avoid, ctx' =
@@ -182,7 +179,7 @@ let pr_context env sigma ctx =
 
 let ppcontext = ppenv_sigma pr_context
 
-let pr_context_map env sigma { src_ctx = delta; map_inst = patcs; tgt_ctx = gamma } =
+let pr_context_map env sigma (delta, patcs, gamma) =
   let env' = push_rel_context delta env in
   let ctx = pr_context env sigma delta in
   let ctx' = pr_context env sigma gamma in
@@ -198,7 +195,7 @@ let ppcontext_map_empty = ppenv_sigma pr_context_map
 
 (** Debugging functions *)
 
-let typecheck_map env evars { src_ctx = ctx; map_inst =  subst; tgt_ctx = ctx' } =
+let typecheck_map env evars (ctx, subst, ctx') =
   typecheck_rel_context env evars ctx;
   typecheck_rel_context env evars ctx';
   let env = push_rel_context ctx env in
@@ -230,12 +227,7 @@ let check_ctx_map ?(unsafe = false) env evars map =
   else map
 
 let mk_ctx_map ?(unsafe = false) env evars ctx subst ctx' =
-  let map = {
-    src_ctx = ctx;
-    map_inst = subst;
-    tgt_ctx = ctx';
-  } in
-  check_ctx_map ~unsafe env evars map
+  let map = (ctx, subst, ctx') in check_ctx_map ~unsafe env evars map
 
 let rec map_patterns f ps =
   List.map (function
@@ -246,11 +238,8 @@ let rec map_patterns f ps =
       | x -> x)
     ps
 
-let map_ctx_map f map = {
-  src_ctx = map_rel_context f map.src_ctx;
-  map_inst = map_patterns f map.map_inst;
-  tgt_ctx = map_rel_context f map.tgt_ctx;
-}
+let map_ctx_map f (g, p, d) =
+  map_rel_context f g, map_patterns f p, map_rel_context f d
 
 (** Specialize by a substitution. *)
 
@@ -296,7 +285,7 @@ let specialize_rel_context sigma s ctx =
       ctx (0, [])
   in res
 
-let mapping_constr sigma (s : context_map) c = specialize_constr sigma s.map_inst c
+let mapping_constr sigma (s : context_map) c = specialize_constr sigma (pi2 s) c
 
 (* Substitute a Constr.t in a pattern. *)
 
@@ -337,11 +326,8 @@ let rec eq_pat env sigma p1 p2 =
   | PInac c, PInac c' -> EConstr.eq_constr sigma c c'
   | _, _ -> false
 
-let make_permutation ?(env = Global.env ()) (sigma : Evd.evar_map) map1 map2 : context_map =
-  let ctx1 = map1.src_ctx in
-  let pats1 = map1.map_inst in
-  let ctx2 = map2.src_ctx in
-  let pats2 = map2.map_inst in
+let make_permutation ?(env = Global.env ()) (sigma : Evd.evar_map)
+    ((ctx1, pats1, _ as map1) : context_map) ((ctx2, pats2, _ as map2) : context_map) : context_map =
   let len = List.length ctx1 in
   let perm = Array.make len None in
   let merge_rels i1 i2 =
@@ -362,7 +348,7 @@ let make_permutation ?(env = Global.env ()) (sigma : Evd.evar_map) map1 map2 : c
     let nenv = Environ.pop_rel_context (Environ.nb_rel env) env in
     let ctx = List.map Context.Rel.Declaration.drop_body (Environ.rel_context env) in
     let nenv = Environ.push_rel_context ctx nenv in
-    let c' = Reductionops.clos_whd_flags RedFlags.all nenv sigma c in
+    let c' = Reductionops.clos_whd_flags CClosure.all nenv sigma c in
     c'
   in
   let env1 = push_rel_context ctx1 env in
@@ -378,7 +364,7 @@ let make_permutation ?(env = Global.env ()) (sigma : Evd.evar_map) map1 map2 : c
         with Invalid_argument _ -> failwith "Could not generate a permutation: different variables"
       else ()
     | PInac c, _ ->
-      let p1' = pat_of_constr env1 sigma (reduce env1 sigma c) in
+      let p1' = pat_of_constr sigma (reduce env1 sigma c) in
       if eq_pat env sigma p1 p1' then
         failwith "Could not generate a permutation: irreducible inaccessible"
       else merge_pats p1' p2
@@ -411,7 +397,7 @@ let make_permutation ?(env = Global.env ()) (sigma : Evd.evar_map) map1 map2 : c
 
   
 let specialize_mapping_constr sigma (m : context_map) c =
-  specialize_constr sigma m.map_inst c
+  specialize_constr sigma (pi2 m) c
 
 let rels_of_ctx ?(with_lets=true) ctx = 
   let len = List.length ctx in
@@ -460,14 +446,14 @@ let rels_above ctx x =
 
 
 
-let is_fix_proto env sigma t =
+let is_fix_proto sigma t =
   match kind sigma t with
-  | LetIn (_, f, _, _) -> is_global env sigma (Lazy.force coq_fix_proto) f
+  | LetIn (_, f, _, _) -> is_global sigma (Lazy.force coq_fix_proto) f
   | _ -> false
 
-let fix_rels env sigma ctx =
+let fix_rels sigma ctx =
   List.fold_left_i (fun i acc decl ->
-      if is_fix_proto env sigma (get_type decl) then Int.Set.add i acc else acc)
+      if is_fix_proto sigma (get_type decl) then Int.Set.add i acc else acc)
     1 Int.Set.empty ctx
 
 let rec dependencies_of_rel ~with_red env evd ctx k x =
@@ -506,7 +492,7 @@ let strengthen ?(full=true) ?(abstract=false) env evd (ctx : rel_context) x (t :
   let rels = 
       Int.Set.union (if full then rels_above ctx x else Int.Set.singleton x)
       (Int.Set.union (dependencies_of_term ~with_red:true env evd ctx t x) 
-      (Int.Set.remove x (fix_rels env evd ctx)))
+      (Int.Set.remove x (fix_rels evd ctx)))
   in
   (* For each variable that we need to push under x, we check
      if its type or body mentions x syntactically. If it does, we normalize
@@ -553,7 +539,7 @@ let new_strengthen (env : Environ.env) (evd : Evd.evar_map) (ctx : rel_context)
     (x : int) ?(rels = rels_above ctx x) (t : constr) :
   context_map * context_map =
   let rels = Int.Set.union rels
-    (Int.Set.union (dependencies_of_term ~with_red:true env evd ctx t x) (fix_rels env evd ctx)) in
+    (Int.Set.union (dependencies_of_term ~with_red:true env evd ctx t x) (fix_rels evd ctx)) in
   let maybe_reduce k t =
     if Int.Set.mem k (Termops.free_rels evd t) then
       Equations_common.nf_betadeltaiota env evd t
@@ -610,7 +596,7 @@ let new_strengthen (env : Environ.env) (evd : Evd.evar_map) (ctx : rel_context)
 
 
 let id_pats g = List.rev (patvars_of_ctx g)
-let id_subst g = { src_ctx = g; map_inst =  id_pats g; tgt_ctx = g }
+let id_subst g = (g, id_pats g, g)
 
 let eq_context_nolet env sigma (g : rel_context) (d : rel_context) =
   try
@@ -636,30 +622,25 @@ let eq_context_nolet env sigma (g : rel_context) (d : rel_context) =
      | e ->
        Printf.eprintf
          "Exception while comparing contexts %s and %s : %s\n"
-         (Pp.string_of_ppcmds (Termops.Internal.print_rel_context (push_rel_context g env) sigma))
-         (Pp.string_of_ppcmds (Termops.Internal.print_rel_context (push_rel_context d env) sigma))
+         (Pp.string_of_ppcmds (Termops.Internal.print_rel_context (push_rel_context g env)))
+         (Pp.string_of_ppcmds (Termops.Internal.print_rel_context (push_rel_context d env)))
          (Printexc.to_string e);
        raise e
 
-let check_eq_context_nolet env sigma snd fst =
-  let g = snd.tgt_ctx in
-  let d = fst.src_ctx in
+let check_eq_context_nolet env sigma (_, _, g as snd) (d, _, _ as fst) =
   if eq_context_nolet env sigma g d then ()
   else errorlabstrm
       (str "Contexts do not agree for composition: "
        ++ pr_context_map env sigma snd ++ str " and " ++ pr_context_map env sigma fst)
 
-let compose_subst ?(unsafe = false) env ?(sigma=Evd.empty) snd fst =
-  let { src_ctx = g; map_inst = p; tgt_ctx = d } = fst in
-  let { src_ctx = g'; map_inst = p'; tgt_ctx = d' } = snd in
+let compose_subst ?(unsafe = false) env ?(sigma=Evd.empty) ((g',p',d') as snd) ((g,p,d) as fst) =
   if !Equations_common.debug && not unsafe then check_eq_context_nolet env sigma snd fst;
   mk_ctx_map ~unsafe env sigma g' (specialize_pats sigma p' p) d
 (*     (g', (specialize_pats p' p), d) *)
 
-let push_mapping_context sigma decl subs =
-  let { src_ctx = g; map_inst = p; tgt_ctx = d } = subs in
+let push_mapping_context sigma decl (g,p,d) =
   let decl' = map_rel_declaration (specialize_constr sigma p) decl in
-  { src_ctx = decl' :: g; map_inst = (PRel 1 :: List.map (lift_pat 1) p); tgt_ctx = decl :: d }
+  (decl' :: g, (PRel 1 :: List.map (lift_pat 1) p), decl :: d)
 
 let lift_subst env evd (ctx : context_map) (g : rel_context) =
   let map = List.fold_right (fun decl acc -> push_mapping_context evd decl acc) g ctx in
@@ -683,7 +664,7 @@ let single_subst ?(unsafe = false) env evd x p g =
         (* 	(List.length g) *)
     in mk_ctx_map ~unsafe env evd substctx pats g
   else
-    let { src_ctx = ctx; map_inst = s; tgt_ctx = g }, invstr = new_strengthen env evd g x t in
+    let (ctx, s, g), invstr = new_strengthen env evd g x t in
     let x' = match nth s (pred x) with PRel i -> i | _ -> error "Occurs check singleton subst"
     and t' = specialize_constr evd s t in
     (* t' is in ctx. Do the substitution of [x'] by [t] now
@@ -700,8 +681,8 @@ let is_local_def i ctx =
   let decl = List.nth ctx (pred i) in
   Context.Rel.Declaration.is_local_def decl
 
-let filter_def_pats map =
+let filter_def_pats (ctx, pats, _) =
   CList.map_filter (function
-      | PRel i when is_local_def i map.src_ctx -> None
-      | PHide i when is_local_def i map.src_ctx -> None
-      | p -> Some p) map.map_inst
+      | PRel i when is_local_def i ctx -> None
+      | PHide i when is_local_def i ctx -> None
+      | p -> Some p) pats
