@@ -211,7 +211,7 @@ val wf_fix_constr :
   Evd.evar_map ref ->
   EConstr.rel_context ->
   EConstr.t ->
-  Univ.Universe.t ->
+  Sorts.t ->
   EConstr.t ->
   EConstr.t ->
   EConstr.t -> EConstr.t * EConstr.t * EConstr.t
@@ -222,7 +222,7 @@ val wf_fix :
   Vars.substl ->
   EConstr.rel_context ->
   EConstr.t ->
-  Univ.Universe.t ->
+  Sorts.t ->
   Constrexpr.constr_expr ->
   Constrexpr.constr_expr option ->
   EConstr.t (* term *) * EConstr.t (* rel *) *
