@@ -353,7 +353,7 @@ let specialize_eqs ~with_block id =
     if worked then
       tclTHENFIRST (Tactics.assert_before_replacing id ty)
         (exact_no_check acc)
-    else tclFAIL 0 (str "Nothing to do in hypothesis " ++ Id.print id ++
+    else tclFAIL (str "Nothing to do in hypothesis " ++ Id.print id ++
                     Printer.pr_econstr_env env !evars ty
                    )
   end
@@ -367,7 +367,7 @@ let specialize_eqs ~with_block id =
   Proofview.Goal.enter begin fun gl ->
   Proofview.tclORELSE (clear [id] <*> Proofview.tclZERO Specialize) begin function
   | (Specialize, _) -> specialize_eqs ~with_block id
-  | e -> tclFAIL 0 (str "Specialization not allowed on dependent hypotheses")
+  | e -> tclFAIL (str "Specialization not allowed on dependent hypotheses")
   end
   end
 
@@ -377,7 +377,7 @@ let dependent_elim_tac ?patterns id : unit Proofview.tactic =
     let env = Proofview.Goal.env gl in
     let concl = Proofview.Goal.concl gl in
     let sigma = Proofview.Goal.sigma gl in
-    let sort = Sorts.univ_of_sort (Retyping.get_sort_of env sigma concl) in
+    let sort = Retyping.get_sort_of env sigma concl in
     let env = Environ.reset_context env in
     let hyps = Proofview.Goal.hyps gl in
     let default_loc, id = id in
