@@ -554,7 +554,7 @@ let autounfold_heads db db' cl =
       match cl with
       | Some hyp -> change_in_hyp ~check:true None (make_change_arg c') hyp
       | None -> convert_concl ~cast:false ~check:false c' DEFAULTcast
-    else tclFAIL 0 (str "Nothing to unfold")
+    else tclFAIL (str "Nothing to unfold")
   end
 
 type hintdb_name = string
@@ -842,8 +842,7 @@ let observe s (tac : unit Proofview.tactic) =
             (Proofview.numgoals >>= fun gls ->
              if gls = 0 then (Feedback.msg_debug (str s ++ str " succeeded"); Proofview.tclUNIT ())
              else Proofview.Goal.enter begin fun gl ->
-              let gl = { Evd.it = Proofview.Goal.goal gl; Evd.sigma = Proofview.Goal.sigma gl } in
-              let () = Feedback.msg_debug (str "Subgoal: " ++ Printer.pr_goal gl) in
+              let () = Feedback.msg_debug (str "Subgoal: " ++ Printer.Debug.pr_goal gl) in
               Proofview.tclUNIT ()
              end))
          (fun iexn -> Feedback.msg_debug
@@ -1075,27 +1074,24 @@ let evd_comb1 f evd x =
 (* Universe related functions *)
 
 let nonalgebraic_universe_level_of_universe env sigma u =
-  match Univ.Universe.level u with
-  | Some l ->
-    if Univ.Level.is_small l then sigma, l, u
-    else
+  match u with
+  | Sorts.Set | Sorts.Prop | Sorts.SProp ->
+    sigma, Univ.Level.set, u
+  | Sorts.Type u0 ->
+    match Univ.Universe.level u0 with
+    | Some l ->
       (match Evd.universe_rigidity sigma l with
       | Evd.UnivFlexible true ->
-        Evd.make_nonalgebraic_variable sigma l, l, Univ.Universe.make l
+        Evd.make_nonalgebraic_variable sigma l, l, Sorts.sort_of_univ @@ Univ.Universe.make l
       | _ -> sigma, l, u)
-  | _ ->
-    let sigma, l = Evd.new_univ_level_variable Evd.univ_flexible sigma in
-    let ul = Univ.Universe.make l in
-    let sigma = Evd.set_leq_sort env sigma (Sorts.sort_of_univ u) (Sorts.sort_of_univ ul) in
-    sigma, l, ul
+    | None ->
+      let sigma, l = Evd.new_univ_level_variable Evd.univ_flexible sigma in
+      let ul = Sorts.sort_of_univ @@ Univ.Universe.make l in
+      let sigma = Evd.set_leq_sort env sigma u ul in
+      sigma, l, ul
 
 let instance_of env sigma ?argu goalu =
   let sigma, goall, goalu = nonalgebraic_universe_level_of_universe env sigma goalu in
-  let sigma, goall, goalu =
-    if Univ.Level.is_prop goall then
-      sigma, Univ.Level.set, Univ.Universe.make Univ.Level.set
-    else sigma, goall, goalu
-  in
   let inst =
     match argu with
     | Some equ ->
