@@ -61,12 +61,12 @@ let inRewRule =
   (* Object just to provide discharge *)
   declare_object
     { (default_object "EQUATIONS_REWRITE_RULE") with
-      cache_function = (fun (_,obj) -> cache_rew_rule obj);
-      discharge_function = (fun (_,x) -> Some x);
+      cache_function = (fun obj -> cache_rew_rule obj);
+      discharge_function = (fun x -> Some x);
       classify_function = (fun _ -> Dispose)
     }
 
-let add_rew_rule ~l2r ~base ref = Lib.add_anonymous_leaf (inRewRule (base,ref,l2r))
+let add_rew_rule ~l2r ~base ref = Lib.add_leaf (inRewRule (base,ref,l2r))
 
 let cache_opacity cst =
   Global.set_strategy (ConstKey cst) Conv_oracle.Opaque
@@ -80,9 +80,9 @@ let inOpacity =
   let obj =
     (* We allow discharging rewrite rules *)
     superglobal_object "EQUATIONS_OPACITY"
-      ~cache:(fun (na, obj) -> cache_opacity obj)
+      ~cache:cache_opacity
       ~subst:(Some subst_opacity)
-      ~discharge:(fun (_, x) -> Some x)
+      ~discharge:(fun x -> Some x)
   in
   declare_object @@ obj
 
@@ -1347,10 +1347,10 @@ let declare_funind ~pm info alias env evd is_rec protos progs
        | Some ((f, _), _, _) -> Global.set_strategy (ConstKey (fst (destConst evd f))) Conv_oracle.transparent)
     else
       ((* Otherwise we turn them opaque and let that information be discharged as well *)
-        Lib.add_anonymous_leaf (inOpacity (fst (destConst evd f)));
+        Lib.add_leaf (inOpacity (fst (destConst evd f)));
         match alias with
         | None -> ()
-        | Some ((f, _), _, _) -> Lib.add_anonymous_leaf (inOpacity (fst (destConst evd f))))
+        | Some ((f, _), _, _) -> Lib.add_leaf (inOpacity (fst (destConst evd f))))
     end;
     pm
   in
@@ -1386,11 +1386,20 @@ let declare_funind ~pm info alias env evd is_rec protos progs
                                 CErrors.print e);
        launch_ind ~pm (Proofview.tclUNIT ())
 
+let max_sort s1 s2 =
+  let open Sorts in
+  match s1, s2 with
+  | (SProp, SProp) | (Prop, Prop) | (Set, Set) -> s1
+  | (SProp, (Prop | Set | Type _ as s)) | ((Prop | Set | Type _) as s, SProp) -> s
+  | (Prop, (Set | Type _ as s)) | ((Set | Type _) as s, Prop) -> s
+  | (Set, Type u) | (Type u, Set) -> Sorts.sort_of_univ (Univ.Universe.sup Univ.Universe.type0 u)
+  | (Type u, Type v) -> Sorts.sort_of_univ (Univ.Universe.sup u v)
+
 let level_of_context env evd ctx acc =
   let _, lev =
     List.fold_right (fun decl (env, lev) ->
         let s = Retyping.get_sort_of env evd (get_type decl) in
-        (push_rel decl env, Univ.sup (Sorts.univ_of_sort s) lev))
+        (push_rel decl env, max_sort s lev))
                     ctx (env,acc)
   in lev
 
@@ -1597,7 +1606,7 @@ let build_equations ~pm with_ind env evd ?(alias:alias option) rec_info progs =
       | Sorts.InProp ->
         (* If the program is producing a proof, then we cannot hope to have its
            graph in Type in general (it might be case-splitting on non-strict propositions). *)
-        Univ.type0m_univ
+        Sorts.prop
       | _ ->
         let ctx = (of_tuple (anonR, None, arity) :: sign) in
         let signlev = level_of_context env !evd ctx sorts in
@@ -1606,14 +1615,14 @@ let build_equations ~pm with_ind env evd ?(alias:alias option) rec_info progs =
     let entry =
       Entries.{ mind_entry_typename = indid;
                 mind_entry_arity = to_constr !evd (it_mkProd_or_LetIn (mkProd (anonR, arity,
-                                                                               mkSort (Sorts.sort_of_univ ind_sort))) sign);
+                                                                               mkSort ind_sort)) sign);
                 mind_entry_consnames = consnames;
                 mind_entry_lc = constructors;
               }
-    in ((entry, sign, arity) :: inds, univs, Univ.sup ind_sort sorts)
+    in ((entry, sign, arity) :: inds, univs, max_sort ind_sort sorts)
   in
   let declare_ind () =
-    let inds, univs, sort = List.fold_left declare_one_ind ([], Univ.Level.Set.empty, Univ.type0m_univ) ind_stmts in
+    let inds, univs, sort = List.fold_left declare_one_ind ([], Univ.Level.Set.empty, Sorts.prop) ind_stmts in
     let sigma = Evd.restrict_universe_context !evd univs in
     let sigma = Evd.minimize_universes sigma in
     let inds =
@@ -1622,7 +1631,7 @@ let build_equations ~pm with_ind env evd ?(alias:alias option) rec_info progs =
                     mind_entry_lc = List.map (to_constr sigma) (List.map of_constr entry.mind_entry_lc);
                     mind_entry_arity =
                       to_constr sigma (it_mkProd_or_LetIn (mkProd (anonR, arity,
-                                                                  mkSort (Sorts.sort_of_univ sort))) sign) })
+                                                                  mkSort sort)) sign) })
         inds
     in
     let univs, ubinders = Evd.univ_entry ~poly sigma in
@@ -1656,7 +1665,7 @@ let build_equations ~pm with_ind env evd ?(alias:alias option) rec_info progs =
                let id = CAst.make @@ scheme in
                (id, false, (kn, i), sort)) 0 inds)
         in
-        Indschemes.do_mutual_induction_scheme ~force_mutual:true mutual;
+        Indschemes.do_mutual_induction_scheme (Global.env()) ~force_mutual:true mutual;
         kn, Smartlocate.global_with_alias (Libnames.qualid_of_ident scheme)
       | _ ->
         let mutual =
@@ -1665,7 +1674,7 @@ let build_equations ~pm with_ind env evd ?(alias:alias option) rec_info progs =
                let id = CAst.make @@ Nameops.add_suffix ind.Entries.mind_entry_typename suff in
                (id, false, (kn, i), sort)) 0 inds)
         in
-        Indschemes.do_mutual_induction_scheme ~force_mutual:true mutual;
+        Indschemes.do_mutual_induction_scheme (Global.env()) ~force_mutual:true mutual;
         let scheme = Nameops.add_suffix (Id.of_string info.base_id) ("_graph" ^ sort_suff) in
         let mutual = List.map2 (fun (i, _, _, _) (_, (_, _, _, _, _, _, _, (kind, cut)), _) ->
                          i, regular_or_nested_rec kind) mutual ind_stmts in
