@@ -121,7 +121,6 @@ let telescope env evd = function
       let (n, _, t) = to_tuple d in
       let len = succ (List.length tl) in
       let ts = Retyping.get_sort_of (push_rel_context tl env) !evd t in
-      let tuniv = Sorts.univ_of_sort ts in
       let ty, tys =
         let rec aux (ty, tyuniv, tys) ds =
           match ds with
@@ -132,18 +131,18 @@ let telescope env evd = function
             let sigty = mkAppG evd (Lazy.force coq_sigma) [|t; pred|] in
             let _, u = destInd !evd (fst (destApp !evd sigty)) in
             let ua = Univ.Instance.to_array (EInstance.kind !evd u) in
-            let l = Univ.Universe.make ua.(0) in
+            let l = Sorts.sort_of_univ @@ Univ.Universe.make ua.(0) in
             let env = push_rel_context ds env in
             (* Ensure that the universe of the sigma is only >= those of t and pred *)
+            let open UnivProblem in
             let enforce_leq env sigma t cstr =
               let ts = Retyping.get_sort_of env sigma t in
-              let su = Sorts.univ_of_sort ts in
-              Univ.enforce_leq su l cstr
+              UnivProblem.Set.add (ULe (ts, l)) cstr
             in
-            let cstrs = enforce_leq env !evd t (Univ.enforce_leq tyuniv l Univ.Constraints.empty) in
-            let () = evd := Evd.add_constraints !evd cstrs in
+            let cstrs = enforce_leq env !evd t (UnivProblem.Set.add (ULe (tyuniv, l)) UnivProblem.Set.empty) in
+            let () = evd := Evd.add_universe_constraints !evd cstrs in
             aux (sigty, l, (u, pred) :: tys) ds
-        in aux (t, tuniv, []) tl
+        in aux (t, ts, []) tl
       in
       let constr, _ = 
 	List.fold_right (fun (u, pred) (intro, k) ->
@@ -341,7 +340,7 @@ let pattern_sigma ~assoc_right c hyp env sigma =
 	    constrs_of_coq_sigma env evd t p @ terms
          | _ -> terms
   in
-  let pat x = Patternops.pattern_of_constr env !evd (to_constr !evd x) in
+  let pat x = Patternops.pattern_of_constr env !evd x in
   let terms = 
     if assoc_right then terms
     else match terms with
@@ -856,7 +855,7 @@ module Tactics =struct
           | None ->
              (tclTHENFIRST (assert_before_replacing id typ)
                            (Logic.refiner ~check:false EConstr.Unsafe.(to_constr prf))))
-      | None -> tclFAIL 0 (str"No currying to do in " ++ Id.print id)
+      | None -> tclFAIL (str"No currying to do in " ++ Id.print id)
   end
 
   let curry =
@@ -868,7 +867,7 @@ module Tactics =struct
       | Prod (na, dom, codom) ->
          Refine.refine ~typecheck:true
            (fun sigma -> curry_concl env sigma na dom codom)
-      | _ -> Tacticals.tclFAIL 0 (str"Goal cannot be curried") end
+      | _ -> Tacticals.tclFAIL (str"Goal cannot be curried") end
 
   let uncurry_call c c' id id' =
     enter begin fun gl ->
@@ -890,7 +889,7 @@ module Tactics =struct
         Proofview.Unsafe.tclEVARS sigma' <*>
         letin_tac None (Name id') (beta_applist sigma (sigpack, [mkVar id])) None nowhere
       with Not_found ->
-        Tacticals.tclFAIL 0 (str"No Signature instance found")
+        Tacticals.tclFAIL (str"No Signature instance found")
     end
 
   let pattern_sigma id =
