@@ -176,7 +176,7 @@ let pr_program_info env sigma p =
   let open Pp in
   Names.Id.print p.program_id ++ str " : " ++
   Printer.pr_econstr_env env sigma (Syntax.program_type p) ++ str " : " ++
-  Printer.pr_econstr_env env sigma (mkType p.program_sort) ++
+  Printer.pr_econstr_env env sigma (mkSort p.program_sort) ++
   str " ( " ++
   (match p.program_rec with
    | Some (Structural ann) ->
@@ -472,7 +472,8 @@ let define_mutual_nested env evd get_prog progs =
   fixes, nested
 
 let helper_evar evm evar env typ src =
-  let sign, typ', instance, _ = push_rel_context_to_named_context ~hypnaming:KeepExistingNames env evm typ in
+  let hypnaming = RenameExistingBut VarSet.full in
+  let sign, typ', instance, _ = push_rel_context_to_named_context ~hypnaming env evm typ in
   let evm' = evar_declare sign evar typ' ~src evm in
     evm', mkEvar (evar, instance)
 
@@ -529,8 +530,8 @@ let term_of_tree env0 isevar sort tree =
 
       (* The next step is to use [simplify]. *)
       let simpl_step = if simpl then
-          Simplify.simplify [None, Simplify.Infer_many] env evd
-        else Simplify.identity env evd
+          Simplify.simplify [None, Simplify.Infer_many]
+        else Simplify.identity
       in
       let branches = Array.map2 (fun (ty, nb, csubst) next ->
         (* We get the context from the constructor arity. *)
@@ -558,7 +559,11 @@ let term_of_tree env0 isevar sort tree =
           msg_debug(str"... named context:");
           msg_debug(Printer.pr_named_context env !evd (EConstr.Unsafe.to_named_context (named_context env)));
         end;
-        let ((hole, c), lsubst) = simpl_step (cut_ctx @ new_ctx @ ctx', ty, sort) in
+        let _ =
+          let env = push_rel_context (cut_ctx @ new_ctx @ ctx') env in
+          evd_comb0 (fun sigma -> Typing.type_of env sigma ty) evd
+        in
+        let ((hole, c), lsubst) = Simplify.apply_simplification_fun simpl_step env evd (cut_ctx @ new_ctx @ ctx', ty, sort) in
         if !debug then
           begin
             let open Feedback in
@@ -962,7 +967,7 @@ let define_one_program_constants flags env0 isevar udecl unfold p =
       let evm', rest' = aux env evm rest in
       let isevar = ref evm' in
       let env = Global.env () in
-      let sort = Sorts.univ_of_sort
+      let sort =
           (Retyping.get_sort_of (push_rel_context (pi1 lhs) env) !isevar info.refined_rettyp) in
       let t, ty = term_of_tree env isevar sort rest' in
       let (cst, (evm, e)) =
@@ -1007,7 +1012,7 @@ type term_info = {
   term_ustate : UState.t;
   base_id : string;
   poly : bool;
-  scope : Locality.locality;
+  scope : Locality.definition_scope;
   decl_kind : Decls.definition_object_kind;
   helpers_info : (Constant.t * (int * int)) list;
   comp_obls : Constant.t list; (** The recursive call proof obligations *)
