@@ -151,7 +151,7 @@ type program_info = {
   program_loc : Loc.t option;
   program_id : Id.t;
   program_orig_type : EConstr.t; (* The original type *)
-  program_sort : Univ.Universe.t; (* The sort of this type *)
+  program_sort : Sorts.t; (* The sort of this type *)
   program_sign : EConstr.rel_context;
   program_arity : EConstr.t;
   program_rec : program_rec_info option;
